import logging
from datetime import datetime
from typing import override
from xml.etree import ElementTree

import httpx
import stamina
from httpx import Client
from lightman_ai.article.models import Article, ArticlesList
from lightman_ai.sources.base import BaseSource
from lightman_ai.sources.exceptions import IncompleteArticleFromSourceError, MalformedSourceResponseError
from pydantic import ValidationError

logger = logging.getLogger("lightman")

_RETRY_ON = httpx.TransportError
_ATTEMPTS = 5
_TIMEOUT = 5


THN_URL = "https://feeds.feedburner.com/TheHackersNews"


class TheHackerNewsSource(BaseSource):
    @override
    def get_articles(self, date: datetime | None = None) -> ArticlesList:
        """Return the articles that are present in THN feed."""
        logger.info("Downloading articles from %s", THN_URL)
        feed = self.get_feed()
        articles = self._xml_to_list_of_articles(feed)
        logger.info("Articles properly downloaded and parsed.")
        if date:
            return ArticlesList.get_articles_from_date_onwards(articles=articles, start_date=date)
        else:
            return ArticlesList(articles=articles)

    def get_feed(self) -> str:
        """Retrieve the TheHackerNews' RSS Feed."""
        for attempt in stamina.retry_context(
            on=_RETRY_ON,
            attempts=_ATTEMPTS,
            timeout=_TIMEOUT,
        ):
            with Client() as http_client, attempt:
                hacker_news_feed = http_client.get(THN_URL)
                hacker_news_feed.raise_for_status()
        return hacker_news_feed.text

    def _xml_to_list_of_articles(self, xml: str) -> list[Article]:
        try:
            root = ElementTree.fromstring(xml)
        except ElementTree.ParseError as e:
            raise MalformedSourceResponseError(f"Invalid XML format: {e}") from e
        channel = root.find("channel")

        if channel is None:
            raise MalformedSourceResponseError("No channel element found in RSS feed")
        items = channel.findall("item")

        parsed = []

        for item in items:
            try:
                title = item.findtext("title", default="").strip()
                description = self._clean(item.findtext("description", default="").strip())
                link = item.findtext("link", default="").strip()
                published_at_str = item.findtext("pubDate", default="").strip()

                if not published_at_str:
                    logger.exception("Missing publication date. link: `%s`", link)
                    raise IncompleteArticleFromSourceError()
                published_at = datetime.strptime(published_at_str, "%a, %d %b %Y %H:%M:%S %z")

                parsed.append(Article(title=title, description=description, link=link, published_at=published_at))
            except (ValidationError, ValueError) as e:
                raise IncompleteArticleFromSourceError from e

        return parsed

    @staticmethod
    def _clean(text: str) -> str:
        """Remove non-useful characters. Helps cleaning the fields that will be sent to the Agent."""
        return text.replace("\\n", "").replace("       ", "")



from collections.abc import Generator
from contextlib import contextmanager
from typing import Any

from lightman_ai.core.exceptions import BaseLightmanError


class BaseGeminiError(BaseLightmanError): ...


class GeminiError(BaseGeminiError): ...


@contextmanager
def map_gemini_exceptions() -> Generator[Any, Any]:
    try:
        yield
    except Exception as err:
        raise GeminiError from err



class BaseLightmanError(Exception): ...


class BaseConfigError(BaseLightmanError): ...


class InvalidConfigError(BaseConfigError): ...


class ConfigNotFoundError(BaseConfigError): ...


class PromptNotFoundError(BaseConfigError): ...



class ArticleBaseError(Exception):
    """Base class for article-related models errors."""


class NoTimeZoneError(ArticleBaseError):
    """Exception class for when there is no timezone associated to a date."""



from abc import ABC, abstractmethod
from datetime import datetime

from lightman_ai.article.models import ArticlesList


class BaseSource(ABC):
    @abstractmethod
    def get_articles(self, date: datetime | None = None) -> ArticlesList: ...



import logging
from collections.abc import AsyncGenerator
from contextlib import asynccontextmanager
from typing import Any

import httpx
from lightman_ai.core.exceptions import BaseLightmanError

logger = logging.getLogger("lightman")


class BaseServiceDeskError(BaseLightmanError):
    """Base exception for all SERVICE_DESK integration errors."""


class ServiceDeskConnectionError(BaseServiceDeskError):
    """Raised when there are network or connection issues with SERVICE_DESK."""


class ServiceDeskAuthenticationError(BaseServiceDeskError):
    """Raised when authentication with SERVICE_DESK fails."""


class ServiceDeskPermissionError(BaseServiceDeskError):
    """Raised when user lacks permissions to perform the requested operation."""


class MissingIssueIDError(BaseServiceDeskError):
    """Error for when we don't get the issue ID in the response."""

    pass


class ServiceDeskHTTPStatusError(BaseServiceDeskError):
    def __init__(self, status_code: int, message: str) -> None:
        self.status_code = status_code
        self.message = message
        super().__init__(f"SERVICE_DESK API error {status_code}: {message}")


class ServiceDeskClientError(ServiceDeskHTTPStatusError):
    """Exception for unmapped HTTP 4xx errors."""

    pass


class ServiceDeskServerError(ServiceDeskHTTPStatusError):
    """Exception for HTTP 5xx errors."""

    pass


class ServiceDeskApiResponseParsingError(BaseServiceDeskError):
    """Raised when SERVICE_DESK API returns an error response that cannot be parsed."""

    def __init__(self, status_code: int, raw_response: str) -> None:
        self.status_code = status_code
        self.raw_response = raw_response
        super().__init__(f"SERVICE_DESK API error {status_code}: Unable to parse error response")


@asynccontextmanager
async def handle_service_desk_exceptions() -> AsyncGenerator[Any, Any]:
    try:
        yield
    except httpx.HTTPStatusError as e:
        if e.response.status_code == httpx.codes.UNAUTHORIZED:
            raise ServiceDeskAuthenticationError from e
        elif e.response.status_code == httpx.codes.FORBIDDEN:
            raise ServiceDeskPermissionError from e
        elif e.response.is_client_error:
            raise ServiceDeskClientError(e.response.status_code, e.response.text) from e
        else:
            raise ServiceDeskServerError(e.response.status_code, e.response.text) from e
    except httpx.TransportError as e:
        raise ServiceDeskConnectionError(f"Network error: {e}") from e



DEFAULT_CONFIG_SECTION = "default"

DEFAULT_CONFIG_FILE = "lightman.toml"

DEFAULT_ENV_FILE = ".env"

DEFAULT_LOG_LEVEL = "WARNING"
VERBOSE_LOG_LEVEL = "INFO"

DEFAULT_AGENT = "openai"
DEFAULT_SCORE = 8
DEFAULT_TIME_ZONE = "UTC"



import logging
from abc import ABC, abstractmethod
from typing import ClassVar, Never, override

from lightman_ai.article.models import SelectedArticlesList
from pydantic_ai import Agent
from pydantic_ai.models.google import GoogleModel
from pydantic_ai.models.openai import OpenAIChatModel


class BaseAgent(ABC):
    _AGENT_CLASS: type[OpenAIChatModel] | type[GoogleModel]
    _DEFAULT_MODEL_NAME: str
    _AGENT_NAME: ClassVar[str]

    def __init__(self, system_prompt: str, model: str | None = None, logger: logging.Logger | None = None) -> None:
        selected_model = model or self._DEFAULT_MODEL_NAME
        agent_model = self._AGENT_CLASS(selected_model)
        self.agent: Agent[Never, SelectedArticlesList] = Agent(
            model=agent_model, output_type=SelectedArticlesList, system_prompt=system_prompt
        )
        self.logger = logger or logging.getLogger("lightman")
        self.logger.info("Selected %s's %s model", self, selected_model)

    @override
    def __str__(self) -> str:
        return self._AGENT_NAME

    @abstractmethod
    def run_prompt(self, prompt: str) -> SelectedArticlesList: ...



class BaseSourceError(Exception):
    """Base Exception class for errors."""


class MalformedSourceResponseError(BaseSourceError):
    """Exception for when the respose format does not match the expected one."""


class IncompleteArticleFromSourceError(MalformedSourceResponseError):
    """Exception for when all the mandatory fields could not be retrieved from an article."""



import math
import re
from collections.abc import Generator
from contextlib import contextmanager
from typing import Any, override

from lightman_ai.core.exceptions import BaseLightmanError
from pydantic_ai.exceptions import ModelHTTPError

from openai import RateLimitError


class BaseOpenAIError(BaseLightmanError): ...


class UnknownOpenAIError(BaseOpenAIError): ...


class OpenAIRateLimitError(BaseOpenAIError):
    regex: str

    @classmethod
    def get_matches(cls, message: str) -> tuple[str, ...]:
        match = re.search(cls.regex, message)
        return match.groups() if match else ()

    @classmethod
    def is_match(cls, message: str) -> bool:
        return bool(re.search(cls.regex, message))


class InputTooLargeError(OpenAIRateLimitError):
    limit: int
    requested: int

    regex: str = (
        r"Limit (\d+), Requested (\d+)\. The input or output tokens must be reduced in order to run successfully\."
    )

    def __init__(self, values: tuple[str, ...]) -> None:
        self.limit = int(values[0])
        self.requested = int(values[1])


class LimitTokensExceededError(OpenAIRateLimitError):
    limit: int
    used: int
    requested: int
    wait_time: int

    regex = r"Limit (\d+), Used (\d+), Requested (\d+)\. Please try again in (\d+\.?(\d+)?)s\."

    def __init__(self, values: tuple[str, ...]) -> None:
        self.limit = int(values[0])
        self.used = int(values[1])
        self.requested = int(values[2])
        self.wait_time = math.ceil(float(values[3]))


class QuotaExceededError(OpenAIRateLimitError):
    regex = r"You exceeded your current quota"

    @override
    @classmethod
    def get_matches(cls, message: str) -> tuple[str, ...]:
        match = re.search(cls.regex, message)
        return match.groups() if match else ()


type TRateLimitErr = type[InputTooLargeError | LimitTokensExceededError]
RATE_LIMIT_ERRORS: list[TRateLimitErr] = [LimitTokensExceededError, InputTooLargeError]


@contextmanager
def map_openai_exceptions() -> Generator[Any, Any]:
    try:
        yield
    except RateLimitError as err:
        for error in RATE_LIMIT_ERRORS:
            if matches := error.get_matches(err.message):
                raise error(matches) from err
        raise UnknownOpenAIError from err
    except ModelHTTPError as err:
        if QuotaExceededError.is_match(err.message):
            raise QuotaExceededError() from err



import logging
import os
from typing import Any, Self

import httpx
import stamina
from lightman_ai.integrations.service_desk.constants import (
    SERVICE_DESK_RETRY_ATTEMPTS,
    SERVICE_DESK_RETRY_ON,
    SERVICE_DESK_RETRY_TIMEOUT,
)
from lightman_ai.integrations.service_desk.exceptions import (
    MissingIssueIDError,
    handle_service_desk_exceptions,
)

logger = logging.getLogger("lightman")


class ServiceDeskIntegration:
    """
    Service Desk integration using httpx to create tickets.

    Provides a simple interface to create Service Desk issues with proper
    authentication and error handling.
    """

    def __init__(
        self,
        base_url: str,
        username: str,
        api_token: str,
        verify_ssl: bool = True,
        timeout: float = 30.0,
    ) -> None:
        self.base_url = base_url.rstrip("/")
        self.username = username
        self.api_token = api_token
        self.client = httpx.AsyncClient(
            base_url=self.base_url,
            auth=(username, api_token),
            headers={
                "Content-Type": "application/json",
                "Accept": "application/json",
            },
            verify=verify_ssl,
            timeout=timeout,
        )

    @classmethod
    def from_env(cls) -> Self:
        required_vars = ["SERVICE_DESK_URL", "SERVICE_DESK_USER", "SERVICE_DESK_TOKEN"]
        missing = [var for var in required_vars if not os.environ.get(var)]
        if missing:
            raise ValueError(f"Missing required environment variable(s): {', '.join(missing)}")
        return cls(
            base_url=os.environ["SERVICE_DESK_URL"],
            username=os.environ["SERVICE_DESK_USER"],
            api_token=os.environ["SERVICE_DESK_TOKEN"],
        )

    async def create_request_of_type(
        self,
        *,
        project_key: str,
        summary: str,
        description: str,
        request_id_type: str,
    ) -> str:
        payload: dict[str, Any] = {
            "serviceDeskId": project_key,
            "requestTypeId": request_id_type,
            "requestFieldValues": {"summary": summary, "description": description},
        }
        for attempt in stamina.retry_context(
            on=SERVICE_DESK_RETRY_ON,
            attempts=SERVICE_DESK_RETRY_ATTEMPTS,
            timeout=SERVICE_DESK_RETRY_TIMEOUT,
        ):
            with attempt:
                async with handle_service_desk_exceptions():
                    response = await self.client.post("/rest/servicedeskapi/request", json=payload)
                    response.raise_for_status()
                    data = response.json()
                    issue_id = data.get("issueId")
                    if not issue_id:
                        raise MissingIssueIDError()
                    logger.info("Successfully created Service Desk issue: %s", issue_id)
        return str(issue_id)



import httpx
from lightman_ai.integrations.service_desk.exceptions import ServiceDeskServerError

# Retry configuration for ServiceDeskIntegration
SERVICE_DESK_RETRY_ON = (httpx.TransportError, httpx.TimeoutException, ServiceDeskServerError)
SERVICE_DESK_RETRY_ATTEMPTS = 3
SERVICE_DESK_RETRY_TIMEOUT = 10



from datetime import date, datetime, time, timedelta
from zoneinfo import ZoneInfo

from lightman_ai.exceptions import MultipleDateSourcesError


def get_start_date(time_zone: str, yesterday: bool, today: bool, start_date: date | None) -> datetime | None:
    mutually_exclusive_date_fields = [x for x in [start_date, today, yesterday] if x]

    if len(mutually_exclusive_date_fields) > 1:
        raise MultipleDateSourcesError(
            "--today, --yesterday and --start-date are mutually exclusive. Set one at a time."
        )

    if today:
        now = datetime.now(ZoneInfo(time_zone))
        return datetime.combine(now, time(0, 0), tzinfo=ZoneInfo(time_zone))
    elif yesterday:
        yesterday_date = datetime.now(ZoneInfo(time_zone)) - timedelta(days=1)
        return datetime.combine(yesterday_date, time(0, 0), tzinfo=ZoneInfo(time_zone))
    elif isinstance(start_date, date):
        return datetime.combine(start_date, time(0, 0), tzinfo=ZoneInfo(time_zone))
    else:
        return None



import time
from typing import override

from lightman_ai.ai.base.agent import BaseAgent
from lightman_ai.ai.openai.exceptions import LimitTokensExceededError, map_openai_exceptions
from lightman_ai.article.models import SelectedArticlesList
from pydantic_ai.agent import AgentRunResult
from pydantic_ai.models.openai import OpenAIChatModel


class OpenAIAgent(BaseAgent):
    """Class that provides an interface to operate with the OpenAI model."""

    _AGENT_CLASS = OpenAIChatModel
    _DEFAULT_MODEL_NAME = "gpt-4.1"
    _AGENT_NAME = "OpenAI"

    def _execute_agent(self, prompt: str) -> AgentRunResult[SelectedArticlesList]:
        with map_openai_exceptions():
            return self.agent.run_sync(prompt)

    @override
    def run_prompt(self, prompt: str) -> SelectedArticlesList:
        try:
            result = self._execute_agent(prompt)
        except LimitTokensExceededError as err:
            self.logger.warning("waiting %s", err.wait_time)
            time.sleep(err.wait_time)
            result = self._execute_agent(prompt)
        return result.output



import logging
import os
from importlib import metadata

logger = logging.getLogger("lightman")


def configure_sentry(log_level: int) -> None:
    """Configure Sentry for error tracking."""
    try:
        import sentry_sdk  # noqa: PLC0415
        from sentry_sdk.integrations.logging import LoggingIntegration  # noqa: PLC0415
    except ImportError:
        if os.getenv("SENTRY_DSN"):
            logger.warning(
                "Could not initialize sentry, it is not installed! Install lightman with `pip install lightman-ai[sentry]` to solve it."
            )
        return

    if not os.getenv("SENTRY_DSN"):
        logger.warning("SENTRY_DSN not configured, skipping Sentry initialization")
        return

    try:
        sentry_logging = LoggingIntegration(level=logging.INFO, event_level=log_level)

        sentry_sdk.init(
            release=metadata.version("lightman-ai"),
            integrations=[sentry_logging],
        )
    except Exception as e:
        logger.warning("Could not instantiate Sentry! %s.\nContinuing with the execution.", e)
